/*
 * abaddri.c - client application interface for abaddrd server
 *
 * Copyright (C) 2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
#include <string.h>	/* memcpy */
#include <sys/types.h>	/* sys/socket.h */
#include <sys/socket.h>	/* AF_INET, etc */
#include "abcidr.h"
#include "abaddr.h"
#include "abaddri.h"

#if !defined(HAVE_STRLCPY)
#define strlcpy(d, s, z) do { strncpy(d, s, z); if ((z) > 0) (d)[(z)-1] = '\0'; } while (0)
#endif

#if !defined(HAVE_STRLCAT)
#define strlcat(d, s, z) do { if ((z) > strlen(d)) strncat(d, s, (z)-strlen(d)-1); } while (0)
#endif

static char *abaddrproc_stat_msgs[] = {
  "abaddrproc: ok",
  "abaddrproc: permission denied",
  "abaddrproc: input/output error",
  "abaddrproc: operation not supported",
  "abaddrproc: unknown error",
};

int abaddrc_init(CLIENT **session, char *host, char **stat_msg)
{
#ifdef TI_RPC
  static char *proto = "datagram_v";
#else
  static char *proto = "udp";
#endif
  CLIENT *clnt;

  if (!(*session = clnt = clnt_create(host, ABADDRPROG, ABADDRVERS, proto))) {
    if (stat_msg) *stat_msg = clnt_spcreateerror(host);
    return -1;
  }
  if (stat_msg) *stat_msg = "abaddrc: connect succeeded";
  return 0;
}
int abaddrc_req_serv(CLIENT *clnt,
		     int rq,
		     char **stat_msg)
{
  abaddrproc_stat *stat;
  abaddr_req arg;

  if (!clnt) {
    if (stat_msg) *stat_msg = "abaddrc: not connected";
    return -1;
  }
  switch (rq) {
  case ABADDR_REQ_DB_LIFE:
  case ABADDR_REQ_DB_TRNC:
  case ABADDR_REQ_DB_LOAD:
  case ABADDR_REQ_DB_SAVE:
    arg = rq;
    break;
  default:
    if (stat_msg) *stat_msg = "abaddrc: illegal server request type";
    return -2;
  }
  if (!(stat = abaddrproc_req_serv_1(&arg, clnt))) {
    if (stat_msg) *stat_msg = clnt_sperror(clnt, "abaddrproc_req_serv_1");
    return -3;
  }
  switch (*stat) {
  case ABADDRPROC_OK:
  case ABADDRPROC_EACCES:
  case ABADDRPROC_EIO:
  case ABADDRPROC_ENOTSUP:
    if (stat_msg) *stat_msg = abaddrproc_stat_msgs[*stat];
    break;
  default:
    if (stat_msg) *stat_msg = abaddrproc_stat_msgs[ABADDRPROC_NG];
  }
  return *stat;
}
int abaddrc_addr_set(CLIENT *clnt,
		     int af, const void *s, const void *e, int rc, long lt, const char *c,
		     char **stat_msg)
{
  abaddrproc_stat *stat;
  abaddrattr arg = { { 0, }, 0, 0, "" };

  if (!clnt) {
    if (stat_msg) *stat_msg = "abaddrc: not connected";
    return -1;
  }
  switch (af) {
  case AF_INET:
    arg.addr.af = ABADDRF_INET;
    *(in_addr_t *)&arg.addr.abaddrs_u.s[0] = ntohl(*(in_addr_t *)s);
    *(in_addr_t *)&arg.addr.abaddrs_u.s[4] = ntohl(*(in_addr_t *)e);
	/* should be host byte order for xdr of rpc */
    break;
#if !defined(DISABLE_IPv6)
  case AF_INET6:
    arg.addr.af = ABADDRF_INET6;
    memcpy(&arg.addr.abaddrs_u.s6[0], s, sizeof(struct in6_addr));
    memcpy(&arg.addr.abaddrs_u.s6[16], e, sizeof(struct in6_addr));
    break;
#endif
  default:
    if (stat_msg) *stat_msg = "abaddrc: not support address family";
    return -2;
  }
  arg.rc = rc;
  arg.lt = lt;
  arg.c = c ? (char *)c : "";
  if (!(stat = abaddrproc_addr_set_1(&arg, clnt))) {
    if (stat_msg) *stat_msg = clnt_sperror(clnt, "abaddrproc_addr_set_1");
    return -3;
  }
  switch (*stat) {
  case ABADDRPROC_OK:
  case ABADDRPROC_EACCES:
  case ABADDRPROC_EIO:
  case ABADDRPROC_ENOTSUP:
    if (stat_msg) *stat_msg = abaddrproc_stat_msgs[*stat];
    break;
  default:
    if (stat_msg) *stat_msg = abaddrproc_stat_msgs[ABADDRPROC_NG];
  }
  return *stat;
}
int abaddrc_addr_get(CLIENT *clnt,
		     int af, const void *s, const void *e,
		     int *f, int *rc, char **stat_msg)
{
  abaddr_get_stat *stat;
  abaddrattr arg = { { 0, }, 0, 0, "" };

  if (!clnt) {
    if (stat_msg) *stat_msg = "abaddrc: not connected";
    return -1;
  }
  switch (af) {
  case AF_INET:
    arg.addr.af = ABADDRF_INET;
    *(in_addr_t *)&arg.addr.abaddrs_u.s[0] = ntohl(*(in_addr_t *)s);
    *(in_addr_t *)&arg.addr.abaddrs_u.s[4] = ntohl(*(in_addr_t *)e);
	/* should be host byte order for xdr of rpc */
    break;
#if !defined(DISABLE_IPv6)
  case AF_INET6:
    arg.addr.af = ABADDRF_INET6;
    memcpy(&arg.addr.abaddrs_u.s6[0], s, sizeof(struct in6_addr));
    memcpy(&arg.addr.abaddrs_u.s6[16], e, sizeof(struct in6_addr));
    break;
#endif
  default:
    if (stat_msg) *stat_msg = "abaddrc: not support address family";
    return -2;
  }
  if (!(stat = abaddrproc_addr_get_1(&arg, clnt))) {	/* ! */
    /*if (stat_msg) *stat_msg = clnt_sperror(clnt, "abaddrproc_addr_get_1");
    return -3;*/
    if (f) *f = AF_UNSPEC;
    return ABADDRPROC_OK;
  }
  switch (stat->stat) {
  case ABADDRPROC_OK: if (f) *f = stat->at.addr.af; if (rc) *rc = stat->at.rc; break;
  default: break;
  }
  switch (stat->stat) {
  case ABADDRPROC_OK:
  case ABADDRPROC_EACCES:
  case ABADDRPROC_EIO:
  case ABADDRPROC_ENOTSUP:
    if (stat_msg) *stat_msg = abaddrproc_stat_msgs[stat->stat];
    break;
  default:
    if (stat_msg) *stat_msg = abaddrproc_stat_msgs[ABADDRPROC_NG];
  }
  return stat->stat;
}
int abaddrc_term(CLIENT *clnt, char **stat_msg)
{
  if (!clnt) {
    if (stat_msg) *stat_msg = "abaddrc: not connected";
    return -1;
  }
  clnt_destroy(clnt);
  if (stat_msg) *stat_msg = "abaddrc: disconnect succeeded";
  return 0;
}
#define no_abaddrd_main
#include "abaddrd.c"

int abaddr_db_wopen(struct abaddr_db *me)
{
  int rv = 0;
  int trunc = me->clear;

  me->clear = 0;
#if defined(USE_QDBM)
  if (!(me->bdb=vlopen(me->pathname, VL_OWRITER|VL_OCREAT|(trunc?VL_OTRUNC:0), abaddr_bdbcmp))) {
    fprintf(stderr, "vlopen failed: %s\n", dperrmsg(dpecode));
    rv = 1;
  }
#else
  if (!((me->bdb=tcbdbnew()) &&
	tcbdbsetcmpfunc(me->bdb, abaddr_bdbcmp, NULL) &&
	tcbdbopen(me->bdb, me->pathname, BDBOWRITER|BDBOCREAT|(trunc?BDBOTRUNC:0)))) {
    fprintf(stderr, "tcbdbopen failed: %s\n", tcbdberrmsg(tcbdbecode(me->bdb)));
    rv = 1;
  }
#endif
  return rv;
}
int abaddr_db_ropen(struct abaddr_db *me)
{
  int rv = 0;

#if defined(USE_QDBM)
  if (!(me->bdb=vlopen(me->pathname, VL_OREADER|VL_ONOLCK, abaddr_bdbcmp))) {
    fprintf(stderr, "vlopen failed: %s\n", dperrmsg(dpecode));
    rv = 1;
  }
#else
  if (!((me->bdb=tcbdbnew()) &&
	tcbdbsetcmpfunc(me->bdb, abaddr_bdbcmp, NULL) &&
	tcbdbopen(me->bdb, me->pathname, BDBOREADER|BDBONOLCK))) {
    fprintf(stderr, "tcbdbopen failed: %s\n", tcbdberrmsg(tcbdbecode(me->bdb)));
    rv = 1;
  }
#endif
  return rv;
}
int abaddr_db_close(struct abaddr_db *me)
{
  int rv = 0;

  abaddr_db_sync(me);
#if defined(USE_QDBM)
  if (!vlclose(me->bdb)) {
    fprintf(stderr, "vlclose failed: %s\n", dperrmsg(dpecode));
    rv = 1;
  }
#else
  if (!tcbdbclose(me->bdb)) {
    fprintf(stderr, "tcbdbclose failed: %s\n", tcbdberrmsg(tcbdbecode(me->bdb)));
    rv = 1;
  }
  tcbdbdel(me->bdb);
#endif
  return rv;
}

int abaddr_req_serv(char *host, char *file,
		    int rq, int argc, char *argv[],
		    char **stat_msg)
{
  int rv = 0;

  if (!host || !host[0]) {
    int check = 0;

#ifdef DEBUG
    check = !0;
#endif
    if (file) me.pathname = file;
    me.argc = argc;
    me.argv = argv;
    switch (rq) {
    case ABADDR_REQ_DB_LIFE:
      if ((rv=abaddr_db_wopen(&me)) != 0) {
	if (stat_msg) *stat_msg = "abaddrc: abaddr_db_wopen failed";
	return rv;
      }
      break;
    case ABADDR_REQ_DB_TRNC:
      me.clear = !0;
      if ((rv=abaddr_db_wopen(&me)) != 0) {
	if (stat_msg) *stat_msg = "abaddrc: abaddr_db_wopen failed";
	return rv;
      }
      break;
    case ABADDR_REQ_DB_LOAD:
      if ((rv=abaddr_db_wopen(&me)) != 0) {
	if (stat_msg) *stat_msg = "abaddrc: abaddr_db_wopen failed";
	return rv;
      }
      break;
    case ABADDR_REQ_DB_SAVE:
      if ((rv=abaddr_db_ropen(&me)) != 0) {
	if (stat_msg) *stat_msg = "abaddrc: abaddr_db_ropen failed";
	return rv;
      }
      break;
    default:
      break;
    }
    switch (rq) {
    case ABADDR_REQ_DB_LIFE:
      abaddr_db_cycle_life(&me);
      break;
    case ABADDR_REQ_DB_TRNC:
      break;
    case ABADDR_REQ_DB_LOAD:
      am_init_masks();
#if !defined(DISABLE_IPv6)
      am6_init_masks();
#endif
      abaddr_db_amset_read(&me,
			   &amset,
#if !defined(DISABLE_IPv6)
			   &am6set,
#endif
			   !0, check);
      break;
    case ABADDR_REQ_DB_SAVE:
      abaddr_db_amset_write(&me);
      break;
    }
    switch (rq) {
    case ABADDR_REQ_DB_LIFE:
    case ABADDR_REQ_DB_TRNC:
    case ABADDR_REQ_DB_LOAD:
    case ABADDR_REQ_DB_SAVE:
      abaddr_db_close(&me);
      break;
    }
  }
  else {
    CLIENT *clnt;

    if ((rv=abaddrc_init(&clnt, host, stat_msg)) != 0)
      return rv;
    if ((rv=abaddrc_req_serv(clnt, rq, stat_msg)) != 0) {
      if (rv != -1) abaddrc_term(clnt, NULL);
      return rv;
    }
    abaddrc_term(clnt, NULL);
  }
  return rv;
}
int abaddr_addr_set(char *host, char *file,
		    int af, const void *s, const void *e, int rc, long lt, const char *c,
		    char **stat_msg)
{
  int rv;

  if (!host || !host[0]) {
    struct abaddrat at;

    if (file) me.pathname = file;
    if ((rv=abaddr_db_wopen(&me)) != 0) {
      if (stat_msg) *stat_msg = "abaddrc: abaddr_db_wopen failed";
      return rv;
    }
    switch ((at.af=af)) {
    case AF_INET:
      at.ads[0].in = *(struct in_addr *)s;
      at.ads[1].in = *(struct in_addr *)e;
      break;
#if !defined(DISABLE_IPv6)
    case AF_INET6:
      at.ads[0].in6 = *(struct in6_addr *)s;
      at.ads[1].in6 = *(struct in6_addr *)e;
      break;
#endif
    default:
      if (stat_msg) *stat_msg = "abaddrc: not support address family";
      return -2;
    }
    at.rc = rc;
    at.lt = lt;
    strlcpy(at.c, c, sizeof(at.c));
    if (!abaddrat_db_del(&me, &at)) {
      if (stat_msg) *stat_msg = "abaddrc: abaddrat_db_del failed";
      rv = -4;
    }
    if (!(at.lt<0))
      if (!abaddrat_db_add(&me, &at)) {
	if (stat_msg) *stat_msg = "abaddrc: abaddrat_db_add failed";
	rv = -5;
      }
    abaddr_db_close(&me);
  }
  else {
    CLIENT *clnt;

    if ((rv=abaddrc_init(&clnt, host, stat_msg)) != 0)
      return rv;
    if ((rv=abaddrc_addr_set(clnt, af, s, e, rc, lt, c, stat_msg)) != 0) {
      if (rv != -1) abaddrc_term(clnt, NULL);
      return rv;
    }
    abaddrc_term(clnt, NULL);
  }
  return rv;
}
int abaddr_addr_get(char *host, char *file,
		    int af, const void *s, const void *e,
		    int *f, int *rc, char **stat_msg)
{
  int rv;

  if (!host || !host[0]) {
    struct abaddrat at;

    if (file) me.pathname = file;
    if ((rv=abaddr_db_ropen(&me)) != 0) {
      if (stat_msg) *stat_msg = "abaddrc: abaddr_db_ropen failed";
      return rv;
    }
    switch ((at.af=af)) {
    case AF_INET:
      at.ads[0].in = *(struct in_addr *)s;
      at.ads[1].in = *(struct in_addr *)e;
      break;
#if !defined(DISABLE_IPv6)
    case AF_INET6:
      at.ads[0].in6 = *(struct in6_addr *)s;
      at.ads[1].in6 = *(struct in6_addr *)e;
      break;
#endif
    default:
      if (f) *f = AF_UNSPEC;
      if (stat_msg) *stat_msg = "abaddrc: not support address family";
      return -2;
    }
    if (f) *f = AF_UNSPEC;
    if (abaddrat_db_get(&me, &at)) {
      if (f) *f = at.af; if (rc) *rc = at.rc;
    }
    abaddr_db_close(&me);
  }
  else {
    CLIENT *clnt;

    if ((rv=abaddrc_init(&clnt, host, stat_msg)) != 0)
      return rv;
    if ((rv=abaddrc_addr_get(clnt, af, s, e, f, rc, stat_msg)) != 0) {
      if (rv != -1) abaddrc_term(clnt, NULL);
      return rv;
    }
    abaddrc_term(clnt, NULL);
  }
  return rv;
}
