/*
 * abaddr.c - utilities for DNSBL in abmail
 *
 * Copyright (C) 2007-2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
/*
 * usage:
 *	$ ./abaddr [-M-] [-P-] 127.0.0.2 ...
 */
#include <stdio.h>	/* stderr, etc */
#include <stdlib.h>	/* qsort, bsearch, etc */
#include <string.h>	/* strcmp, etc */
#include <sys/types.h>	/* netinet/in.h, arpa/inet.h, sys/socket.h, etc */
#include <sys/socket.h>	/* AF_INET6, etc */
#include <netinet/in.h>	/* in_addr, etc */
#include <arpa/inet.h>	/* inet_ntoa, etc */
#include <netdb.h>	/* gethostbyname, etc */
#if !defined(DISABLE_PTHREAD)
#include <pthread.h>	/* pthread_create, etc */
#endif

#if !defined(HAVE_STRLCPY)
#define strlcpy(d, s, z) do { strncpy(d, s, z); if ((z) > 0) (d)[(z)-1] = '\0'; } while (0)
#endif

#if !defined(HAVE_INET_ATON)
#define inet_aton(s, p) ((((p)->s_addr=inet_addr(s))!=-1||strcmp(s, "255.255.255.255")==0)?1:0)
#endif

char *lookup_reverse(const char *a, char *b, size_t z, int paranoid)
{
  char *rv = NULL, s[2048];
#if !defined(DISABLE_PTHREAD)
  struct addrinfo hints, *res, *r;
  int err, flag = 0;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags |= AI_NUMERICHOST;
  if (!(err=getaddrinfo(a, NULL, &hints, &res))) {
    for (r=res; r!=0; r=r->ai_next) {
      if (!(err=getnameinfo(r->ai_addr, r->ai_addrlen, s, sizeof(s),
			    NULL, 0, NI_NAMEREQD))) {
	if (!paranoid) {
	  strlcpy(b, s, z);
	  rv = b;
	  break;
	}
	else {
	  struct addrinfo hints, *res, *q;
	  int err;

	  memset(&hints, 0, sizeof(hints));
	  hints.ai_family = AF_UNSPEC;
	  hints.ai_socktype = SOCK_STREAM;
	  if (!(err=getaddrinfo(s, NULL, &hints, &res))) {
	    for (q=res; q!=0; q=q->ai_next) {
	      if (r->ai_family == AF_INET && r->ai_family == q->ai_family &&
		  ((struct sockaddr_in *)r->ai_addr)->sin_addr.s_addr ==
		  ((struct sockaddr_in *)q->ai_addr)->sin_addr.s_addr) {
		flag = !0;
		break;
	      }
#if !defined(DISABLE_IPv6)
	      if (r->ai_family == AF_INET6 && r->ai_family == q->ai_family &&
		  memcmp(&((struct sockaddr_in6 *)r->ai_addr)->sin6_addr,
			 &((struct sockaddr_in6 *)q->ai_addr)->sin6_addr,
			 sizeof(struct in6_addr)) == 0) {
		flag = !0;
		break;
	      }
#endif
	    }
	    freeaddrinfo(res);
	  }
	  if (flag) {
	    strlcpy(b, s, z);
	    rv = b;
	    break;
	  }
	}
      }
    }
    freeaddrinfo(res);
  }
#else
  struct hostent *he;
  struct in_addr in;
#if !defined(DISABLE_IPv6)
  struct in6_addr in6;
#endif

  in.s_addr = -1;
  if ((inet_aton(a, &in) != 0 &&
       (he=gethostbyaddr((char *)&in, sizeof(in), AF_INET)))
#if !defined(DISABLE_IPv6)
      ||
      (inet_pton(AF_INET6, a, &in6) != 0 &&
       (he=gethostbyaddr((char *)&in6, sizeof(in6), AF_INET6)))
#endif
      ) {
    strlcpy(s, he->h_name, sizeof(s));
    if (!paranoid) {
      strlcpy(b, s, z);
      rv = b;
    }
    else {
      struct hostent *he;
      int i, flag = 0;

      if (in.s_addr != -1 && (he=gethostbyname(s)))
	for (i=0; he->h_addr_list[i]; i++)
	  if (in.s_addr == (*(struct in_addr *)he->h_addr_list[i]).s_addr) {
	    flag = !0;
	    break;
	  }
#if !defined(DISABLE_IPv6)
      if (!flag && (he=gethostbyname2(s, AF_INET6)))
	for (i=0; he->h_addr_list[i]; i++)
	  if (memcmp(in6.s6_addr,
		     ((struct in6_addr *)he->h_addr_list[i])->s6_addr,
		     sizeof(struct in6_addr)) == 0) {
	    flag = !0;
	    break;
	  }
#endif
      if (flag) {
	strlcpy(b, s, z);
	rv = b;
      }
    }
  }
#endif
  return rv;
}

const char *dnsbls[] = {
  NULL,
  /*"list.dsbl.org",*/		/* expired */
  /*"unconfirmed.dsbl.org",*/	/* expired */
  /*"spam.dnsbl.sorbs.net",*/	/* too unsparing */
  "cbl.abuseat.org",
  /*"opm.blitzed.org",*/	/* expired */
  /*"dnsbl.sorbs.net",*/	/* too unsparing */
  "bl.spamcop.net",
  /*"spam.tqmcube.com",*/	/* expired, unsparing */
  "bl.spamcannibal.org",	/* too unsparing */
  "dnsbl-1.uceprotect.net",
  "dnsbl-2.uceprotect.net",
  "dnsbl-3.uceprotect.net",
  "zen.spamhaus.org",
  //"niku.2ch.net",
};

#if !defined(DISABLE_PTHREAD)
struct thr_dnsbl_find_arg {
  char name[2048];
  int i, *rv, *n;
  pthread_t *selves;
  pthread_mutex_t *mutex;
};

int thr_dnsbl_find_enable = !0;

void *thr_dnsbl_find(void *arg)
{
  int rv = 0;
  struct thr_dnsbl_find_arg *a = (struct thr_dnsbl_find_arg *)arg;
  struct addrinfo hints, *res;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  if ((rv = (getaddrinfo(a->name, NULL, &hints, &res) == 0)?a->i:0)) {
    freeaddrinfo(res);
    if (*a->rv == 0) {
      pthread_mutex_lock(a->mutex);
      *a->rv = rv;
      pthread_mutex_unlock(a->mutex);
    }
  }
  return NULL;
}
#endif

int dnsbl_find(const char *ip)
{
  int rv = 0;
  int i;
  unsigned int ip4[4];

#if !defined(DISABLE_PTHREAD)
  if (!thr_dnsbl_find_enable) {
#endif
    struct hostent *he;
    char name[2048];

    if (sscanf(ip, "%u.%u.%u.%u", &ip4[0], &ip4[1], &ip4[2], &ip4[3]) != 4)
      return rv;
    for (i=1; i<sizeof(dnsbls)/sizeof(char *); i++) {
      sprintf(name, "%u.%u.%u.%u.%s",
	      ip4[3], ip4[2], ip4[1], ip4[0], dnsbls[i]);	/* not insecure */
      if ((he=gethostbyname(name))) {
	rv = i;
	break;
      }
    }
#if !defined(DISABLE_PTHREAD)
  }
  else {
    int n = 1;
    pthread_t threads[sizeof(dnsbls)/sizeof(char *)];
    pthread_mutex_t mutex;
    struct thr_dnsbl_find_arg args[sizeof(dnsbls)/sizeof(char *)];

    if (sscanf(ip, "%u.%u.%u.%u", &ip4[0], &ip4[1], &ip4[2], &ip4[3]) != 4)
      return rv;
    pthread_mutex_init(&mutex, NULL);
    for (i=1; i<sizeof(dnsbls)/sizeof(char *) && rv == 0; i++) {
      snprintf(args[i].name, sizeof(args[i].name), "%u.%u.%u.%u.%s",
	       ip4[3], ip4[2], ip4[1], ip4[0], dnsbls[i]);
      args[i].i = i;
      args[i].rv = &rv;
      args[i].n = &n;
      args[i].selves = threads;
      args[i].mutex = &mutex;
      pthread_create(&threads[i], NULL, thr_dnsbl_find, (void *)&args[i]);
      n++;
    }
    if (rv) {
      pthread_join(threads[rv], NULL);
      for (i=1; i<n; i++)
	if (i != rv) pthread_detach(threads[i]);
    }
    else
      for (i=1; i<n; i++)
	pthread_join(threads[i], NULL);
  }
#endif
  return rv;
}

int main(int argc, char *argv[])
{
  int rv, a, paranoid = !0;

  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "--") == 0)
      break;
    else if (strcmp(argv[a], "-M-") == 0)
#if !defined(DISABLE_PTHREAD)
      thr_dnsbl_find_enable = 0
#endif
	;
    else if (strcmp(argv[a], "-P-") == 0)
      paranoid = 0;
    else
      break;

  for (; a<argc; a++) {
    char s[2048], *r;

    printf("hostname: %s\n",
	   (r=lookup_reverse(argv[a], s, sizeof(s), paranoid)) ? r : "unknown");
    if ((rv = dnsbl_find(argv[a])))
      printf("%s:\t%s\n", argv[a], dnsbls[rv]);
  }
  return 0;
}
