/*
 * ararray.hpp - arithmetic-array class template for a fixed multisize integer
 *
 * Copyright (C) 2009 Taiji Yamada <taiji@aihara.co.jp>
 *
 * Distributed under the Boost Software License, Version 1.0. See
 * http://www.boost.org/LICENSE_1_0.txt for more details.
 */
#ifndef _ararray_hpp_
#define _ararray_hpp_
#include <arr_ar.h>
#include <ostream> // not implemented yet

template<typename T, const int n>
class ararray {
  T a[n];
public:
  ararray()
  {
    arr_each(n, a[i] = 0);//memset(a, 0, sizeof(a));
    // ref. H. Sutter & A. Alexandrescu, ``C++ Conding Standards,''
    // Item 96, 2005.
  }
  ararray(unsigned long long v)
  {
    set_ull(v);
  }
  ararray(T a[n])
  {
    arr_each(n, this->a[i] = a[i]);//memcpy(this->a, a, sizeof(this->a));
    // ref. H. Sutter & A. Alexandrescu, ``C++ Conding Standards,''
    // Item 96, 2005.
  }

  // postfix increment and decrement operators
  // ref. S. Meyers, ``More Effective C++,'' Item 6, 1996.
  const ararray operator++(int)
  {
    ararray<T, n> r = *this;
    ++(*this);
    return r;
  }
  const ararray operator--(int)
  {
    ararray<T, n> r = *this;
    --(*this);
    return r;
  }
  // prefix increment and decrement operators
  ararray &operator++()
  {
#if 0 /* simple version for increment */
    ararray<T, n> b(1), c;
    arr_add(n, a, b.a, c.a);
#else
    ararray<T, n> c;
    arr_inc(n, a, c.a);
#endif
    arr_each(n, a[i] = c.a[i]);//memcpy(a, c.a, sizeof(a));
    return *this;
  }
  ararray &operator--()
  {
#if 0 /* simple version for decrement */
    ararray<T, n> b(1), c;
    arr_sub(n, a, b.a, c.a);
#else
    ararray<T, n> c;
    arr_dec(n, a, c.a);
#endif
    arr_each(n, a[i] = c.a[i]);//memcpy(a, c.a, sizeof(a));
    return *this;
  }

  const bool operator!() const
  {
    return !get_b();
  }
  // ref. S. Meyers, ``Effective C++, 3rd ed.,'' Item 3, 2005.
  const ararray operator~() const
  {
    ararray<T, n> r;
    arr_not(n, a, r.a);
    return r;
  }

  const ararray operator+(const ararray b) const
  {
    ararray<T, n> c;
    arr_add(n, a, b.a, c.a);
    return c;
  }
  const ararray operator-(const ararray b) const
  {
    ararray<T, n> c;
    arr_sub(n, a, b.a, c.a);
    return c;
  }

  const ararray operator<<(const unsigned int s) const
  {
    ararray<T, n> r;
    arr_lsl(n, a, s, r.a);
    return r;
  }
  const ararray operator>>(const unsigned int s) const
  {
    ararray<T, n> r;
    arr_lsr(n, a, s, r.a);
    return r;
  }

  const ararray operator&(const ararray b) const
  {
    ararray<T, n> c;
    arr_and(n, a, b.a, c.a);
    return c;
  }
  const ararray operator^(const ararray b) const
  {
    ararray<T, n> c;
    arr_xor(n, a, b.a, c.a);
    return c;
  }
  const ararray operator|(const ararray b) const
  {
    ararray<T, n> c;
    arr_or(n, a, b.a, c.a);
    return c;
  }

  // ref. S. Meyers, ``Effective C++, 3rd ed.,'' Item 48, 2005.
  ararray &operator+=(const ararray &b)
  {
    ararray<T, n> c;
    arr_add(n, a, b.a, c.a);
    arr_each(n, a[i] = c.a[i]);//memcpy(a, c.a, sizeof(a));
    return *this;
  }
  ararray &operator-=(const ararray &b)
  {
    ararray<T, n> c;
    arr_sub(n, a, b.a, c.a);
    arr_each(n, a[i] = c.a[i]);//memcpy(a, c.a, sizeof(a));
    return *this;
  }
  ararray &operator<<=(const unsigned int s)
  {
    arr_lsl(n, a, s, a);
    return *this;
  }
  ararray &operator>>=(const unsigned int s)
  {
    arr_lsr(n, a, s, a);
    return *this;
  }

  ararray &operator&=(const ararray &b)
  {
    arr_and(n, a, b.a, a);
    return *this;
  }
  ararray &operator^=(const ararray &b)
  {
    arr_xor(n, a, b.a, a);
    return *this;
  }
  ararray &operator|=(const ararray &b)
  {
    arr_or(n, a, b.a, a);
    return *this;
  }

  const int cmp(const ararray &b) const
  {
    int cmp;
    arr_is(n, (cmp=(a[i]>b.a[i]?1:a[i]<b.a[i]?-1:0)) != 0);
    return cmp;
  }
  const bool operator<(const ararray &b) const
  {
    return cmp(b) == -1;
  }
  const bool operator<=(const ararray &b) const
  {
    return cmp(b) != 1;
  }
  const bool operator>(const ararray &b) const
  {
    return cmp(b) == 1;
  }
  const bool operator>=(const ararray &b) const
  {
    return cmp(b) != -1;
  }
  const bool operator==(const ararray &b) const
  {
    return cmp(b) == 0;
  }
  const bool operator!=(const ararray &b) const
  {
    return cmp(b) != 0;
  }

  // ref. S. Meyers, ``More Effective C++,'' Item 5, 1996.
  const bool get_b() const
  {
    bool r;
    arr_is_rev(n, (r=(a[i]!=0)));
    return r;
  }
  const unsigned long get_ul() const
  {
    int i;
    unsigned long r = 0;
    for (i=n-1; i>=0; i--) {
      const unsigned long m = 1UL << (n-1-i)*8*sizeof(T);
      r += m*a[i];
    }
    return r;
  }
  const unsigned long long get_ull() const
  {
    int i;
    unsigned long long r = 0;
    for (i=n-1; i>=0; i--) {
      unsigned long long m = 1ULL << (n-1-i)*8*sizeof(T);
      r += m*a[i];
    }
    return r;
  }
  const double get_d() const
  {
    int i;
    double r = 0;
    for (i=n-1; i>=0; i--) {
      const double m = 1ULL << (n-1-i)*8*sizeof(T);
      r += m*a[i];
    }
    return r;
  }
  void set_ul(unsigned long v)
  {
    int i;
    for (i=0; i<n && v; i++) {
      const unsigned long m = 1UL << (n-1-i)*8*sizeof(T);
      if ((a[i] = v/m))
	v %= m;
    }
    for (; i<n; i++)
      a[i] = 0;
  }
  void set_ull(unsigned long long v)
  {
    int i;
    for (i=0; i<n && v; i++) {
      const unsigned long long m = 1ULL << (n-1-i)*8*sizeof(T);
      if ((a[i] = v/m))
	v %= m;
    }
    for (; i<n; i++)
      a[i] = 0;
  }
  void set_d(double v)
  {
    int i;
    for (i=0; i<n && v; i++) {
      const double m = 1ULL << (n-1-i)*8*sizeof(T);
      double d;
      if (!((d=v/m) < 1)) {
	a[i] = d;
	v -= a[i]*m;
      }
    }
    for (; i<n; i++)
      a[i] = 0;
  }
};

template<typename T, const int n>
inline std::ostream &operator<<(std::ostream &o, const ararray<T, n> &a)
{
  return o << a.get_d(); // not implemented yet
}
#endif
